// Copyright (C) MongoDB, Inc. 2017-present.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License. You may obtain
// a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

// DO NOT EDIT. Code generated by Github action
// "mongodb-labs/drivers-github-tools/golang/pre-publish".

// Package version defines the Go Driver version.
package version

// Driver is the current version of the driver.
var Driver = "1.17.6"
